package org.apache.bcel.generic;

import org.apache.bcel.Const;
import org.apache.bcel.classfile.Constant;
import java.util.NoSuchElementException;
import java.util.Iterator;
import org.apache.commons.lang3.ArrayUtils;
import java.io.OutputStream;
import java.io.DataOutputStream;
import java.io.ByteArrayOutputStream;
import java.util.Map;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Arrays;
import java.io.IOException;
import org.apache.bcel.util.ByteSequence;
import java.util.List;

public class InstructionList implements Iterable<InstructionHandle>
{
    private InstructionHandle start;
    private InstructionHandle end;
    private int length;
    private int[] bytePositions;
    private List<InstructionListObserver> observers;
    
    public static InstructionHandle findHandle(final InstructionHandle[] ihs, final int[] pos, final int count, final int target) {
        int l = 0;
        int r = count - 1;
        do {
            final int i = l + r >>> 1;
            final int j = pos[i];
            if (j == target) {
                return ihs[i];
            }
            if (target < j) {
                r = i - 1;
            }
            else {
                l = i + 1;
            }
        } while (l <= r);
        return null;
    }
    
    public InstructionList() {
    }
    
    public InstructionList(final BranchInstruction i) {
        this.append(i);
    }
    
    public InstructionList(final byte[] code) {
        int count = 0;
        InstructionHandle[] ihs;
        int[] pos;
        try (ByteSequence bytes = new ByteSequence(code)) {
            ihs = new InstructionHandle[code.length];
            pos = new int[code.length];
            while (bytes.available() > 0) {
                final int off = bytes.getIndex();
                pos[count] = off;
                final Instruction i = Instruction.readInstruction(bytes);
                InstructionHandle ih;
                if (i instanceof BranchInstruction) {
                    ih = this.append((BranchInstruction)i);
                }
                else {
                    ih = this.append(i);
                }
                ih.setPosition(off);
                ihs[count] = ih;
                ++count;
            }
        }
        catch (final IOException e) {
            throw new ClassGenException(e.toString(), e);
        }
        this.bytePositions = Arrays.copyOf(pos, count);
        for (int j = 0; j < count; ++j) {
            if (ihs[j] instanceof BranchHandle) {
                final BranchInstruction bi = (BranchInstruction)ihs[j].getInstruction();
                int target = bi.getPosition() + bi.getIndex();
                InstructionHandle ih2 = findHandle(ihs, pos, count, target);
                if (ih2 == null) {
                    throw new ClassGenException("Couldn't find target for branch: " + bi);
                }
                bi.setTarget(ih2);
                if (bi instanceof Select) {
                    final Select s = (Select)bi;
                    final int[] indices = s.getIndices();
                    for (int k = 0; k < indices.length; ++k) {
                        target = bi.getPosition() + indices[k];
                        ih2 = findHandle(ihs, pos, count, target);
                        if (ih2 == null) {
                            throw new ClassGenException("Couldn't find target for switch: " + bi);
                        }
                        s.setTarget(k, ih2);
                    }
                }
            }
        }
    }
    
    public InstructionList(final CompoundInstruction c) {
        this.append(c.getInstructionList());
    }
    
    public InstructionList(final Instruction i) {
        this.append(i);
    }
    
    public void addObserver(final InstructionListObserver o) {
        if (this.observers == null) {
            this.observers = new ArrayList<InstructionListObserver>();
        }
        this.observers.add(o);
    }
    
    public BranchHandle append(final BranchInstruction i) {
        final BranchHandle ih = BranchHandle.getBranchHandle(i);
        this.append(ih);
        return ih;
    }
    
    public InstructionHandle append(final CompoundInstruction c) {
        return this.append(c.getInstructionList());
    }
    
    public InstructionHandle append(final Instruction i) {
        final InstructionHandle ih = InstructionHandle.getInstructionHandle(i);
        this.append(ih);
        return ih;
    }
    
    public InstructionHandle append(final Instruction i, final CompoundInstruction c) {
        return this.append(i, c.getInstructionList());
    }
    
    public InstructionHandle append(final Instruction i, final Instruction j) {
        return this.append(i, new InstructionList(j));
    }
    
    public InstructionHandle append(final Instruction i, final InstructionList il) {
        final InstructionHandle ih;
        if ((ih = this.findInstruction2(i)) == null) {
            throw new ClassGenException("Instruction " + i + " is not contained in this list.");
        }
        return this.append(ih, il);
    }
    
    private void append(final InstructionHandle ih) {
        if (this.isEmpty()) {
            this.end = ih;
            (this.start = ih).setNext(ih.setPrev(null));
        }
        else {
            this.end.setNext(ih);
            ih.setPrev(this.end);
            ih.setNext(null);
            this.end = ih;
        }
        ++this.length;
    }
    
    public BranchHandle append(final InstructionHandle ih, final BranchInstruction i) {
        final BranchHandle bh = BranchHandle.getBranchHandle(i);
        final InstructionList il = new InstructionList();
        il.append(bh);
        this.append(ih, il);
        return bh;
    }
    
    public InstructionHandle append(final InstructionHandle ih, final CompoundInstruction c) {
        return this.append(ih, c.getInstructionList());
    }
    
    public InstructionHandle append(final InstructionHandle ih, final Instruction i) {
        return this.append(ih, new InstructionList(i));
    }
    
    public InstructionHandle append(final InstructionHandle ih, final InstructionList il) {
        if (il == null) {
            throw new ClassGenException("Appending null InstructionList");
        }
        if (il.isEmpty()) {
            return ih;
        }
        final InstructionHandle next = ih.getNext();
        final InstructionHandle ret = il.start;
        ih.setNext(il.start);
        il.start.setPrev(ih);
        il.end.setNext(next);
        if (next != null) {
            next.setPrev(il.end);
        }
        else {
            this.end = il.end;
        }
        this.length += il.length;
        il.clear();
        return ret;
    }
    
    public InstructionHandle append(final InstructionList il) {
        if (il == null) {
            throw new ClassGenException("Appending null InstructionList");
        }
        if (il.isEmpty()) {
            return null;
        }
        if (this.isEmpty()) {
            this.start = il.start;
            this.end = il.end;
            this.length = il.length;
            il.clear();
            return this.start;
        }
        return this.append(this.end, il);
    }
    
    private void clear() {
        final InstructionHandle instructionHandle = null;
        this.end = instructionHandle;
        this.start = instructionHandle;
        this.length = 0;
    }
    
    public boolean contains(final Instruction i) {
        return this.findInstruction1(i) != null;
    }
    
    public boolean contains(final InstructionHandle i) {
        if (i == null) {
            return false;
        }
        for (InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
            if (ih == i) {
                return true;
            }
        }
        return false;
    }
    
    public InstructionList copy() {
        final Map<InstructionHandle, InstructionHandle> map = new HashMap<InstructionHandle, InstructionHandle>();
        final InstructionList il = new InstructionList();
        for (InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
            final Instruction i = ih.getInstruction();
            final Instruction c = i.copy();
            if (c instanceof BranchInstruction) {
                map.put(ih, il.append((BranchInstruction)c));
            }
            else {
                map.put(ih, il.append(c));
            }
        }
        for (InstructionHandle ih = this.start, ch = il.start; ih != null; ih = ih.getNext(), ch = ch.getNext()) {
            final Instruction j = ih.getInstruction();
            final Instruction c2 = ch.getInstruction();
            if (j instanceof BranchInstruction) {
                final BranchInstruction bi = (BranchInstruction)j;
                final BranchInstruction bc = (BranchInstruction)c2;
                final InstructionHandle itarget = bi.getTarget();
                bc.setTarget(map.get(itarget));
                if (bi instanceof Select) {
                    final InstructionHandle[] itargets = ((Select)bi).getTargets();
                    final InstructionHandle[] ctargets = ((Select)bc).getTargets();
                    for (int k = 0; k < itargets.length; ++k) {
                        ctargets[k] = map.get(itargets[k]);
                    }
                }
            }
        }
        return il;
    }
    
    public void delete(final Instruction i) throws TargetLostException {
        final InstructionHandle ih;
        if ((ih = this.findInstruction1(i)) == null) {
            throw new ClassGenException("Instruction " + i + " is not contained in this list.");
        }
        this.delete(ih);
    }
    
    public void delete(final Instruction from, final Instruction to) throws TargetLostException {
        final InstructionHandle fromIh;
        if ((fromIh = this.findInstruction1(from)) == null) {
            throw new ClassGenException("Instruction " + from + " is not contained in this list.");
        }
        final InstructionHandle toIh;
        if ((toIh = this.findInstruction2(to)) == null) {
            throw new ClassGenException("Instruction " + to + " is not contained in this list.");
        }
        this.delete(fromIh, toIh);
    }
    
    public void delete(final InstructionHandle ih) throws TargetLostException {
        this.remove(ih.getPrev(), ih.getNext());
    }
    
    public void delete(final InstructionHandle from, final InstructionHandle to) throws TargetLostException {
        this.remove(from.getPrev(), to.getNext());
    }
    
    public void dispose() {
        for (InstructionHandle ih = this.end; ih != null; ih = ih.getPrev()) {
            ih.dispose();
        }
        this.clear();
    }
    
    public InstructionHandle findHandle(final int pos) {
        final int[] positions = this.bytePositions;
        InstructionHandle ih = this.start;
        for (int i = 0; i < this.length; ++i) {
            if (positions[i] == pos) {
                return ih;
            }
            ih = ih.getNext();
        }
        return null;
    }
    
    private InstructionHandle findInstruction1(final Instruction i) {
        for (InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
            if (ih.getInstruction() == i) {
                return ih;
            }
        }
        return null;
    }
    
    private InstructionHandle findInstruction2(final Instruction i) {
        for (InstructionHandle ih = this.end; ih != null; ih = ih.getPrev()) {
            if (ih.getInstruction() == i) {
                return ih;
            }
        }
        return null;
    }
    
    public byte[] getByteCode() {
        this.setPositions();
        final ByteArrayOutputStream b = new ByteArrayOutputStream();
        final DataOutputStream out = new DataOutputStream(b);
        try {
            for (InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
                final Instruction i = ih.getInstruction();
                i.dump(out);
            }
            out.flush();
        }
        catch (final IOException e) {
            System.err.println(e);
            return ArrayUtils.EMPTY_BYTE_ARRAY;
        }
        return b.toByteArray();
    }
    
    public InstructionHandle getEnd() {
        return this.end;
    }
    
    public InstructionHandle[] getInstructionHandles() {
        final InstructionHandle[] ihs = new InstructionHandle[this.length];
        InstructionHandle ih = this.start;
        for (int i = 0; i < this.length; ++i) {
            ihs[i] = ih;
            ih = ih.getNext();
        }
        return ihs;
    }
    
    public int[] getInstructionPositions() {
        return this.bytePositions;
    }
    
    public Instruction[] getInstructions() {
        final List<Instruction> instructions = new ArrayList<Instruction>();
        try (ByteSequence bytes = new ByteSequence(this.getByteCode())) {
            while (bytes.available() > 0) {
                instructions.add(Instruction.readInstruction(bytes));
            }
        }
        catch (final IOException e) {
            throw new ClassGenException(e.toString(), e);
        }
        return instructions.toArray(Instruction.EMPTY_ARRAY);
    }
    
    public int getLength() {
        return this.length;
    }
    
    public InstructionHandle getStart() {
        return this.start;
    }
    
    public BranchHandle insert(final BranchInstruction i) {
        final BranchHandle ih = BranchHandle.getBranchHandle(i);
        this.insert(ih);
        return ih;
    }
    
    public InstructionHandle insert(final CompoundInstruction c) {
        return this.insert(c.getInstructionList());
    }
    
    public InstructionHandle insert(final Instruction i) {
        final InstructionHandle ih = InstructionHandle.getInstructionHandle(i);
        this.insert(ih);
        return ih;
    }
    
    public InstructionHandle insert(final Instruction i, final CompoundInstruction c) {
        return this.insert(i, c.getInstructionList());
    }
    
    public InstructionHandle insert(final Instruction i, final Instruction j) {
        return this.insert(i, new InstructionList(j));
    }
    
    public InstructionHandle insert(final Instruction i, final InstructionList il) {
        final InstructionHandle ih;
        if ((ih = this.findInstruction1(i)) == null) {
            throw new ClassGenException("Instruction " + i + " is not contained in this list.");
        }
        return this.insert(ih, il);
    }
    
    private void insert(final InstructionHandle ih) {
        if (this.isEmpty()) {
            this.end = ih;
            (this.start = ih).setNext(ih.setPrev(null));
        }
        else {
            this.start.setPrev(ih);
            ih.setNext(this.start);
            ih.setPrev(null);
            this.start = ih;
        }
        ++this.length;
    }
    
    public BranchHandle insert(final InstructionHandle ih, final BranchInstruction i) {
        final BranchHandle bh = BranchHandle.getBranchHandle(i);
        final InstructionList il = new InstructionList();
        il.append(bh);
        this.insert(ih, il);
        return bh;
    }
    
    public InstructionHandle insert(final InstructionHandle ih, final CompoundInstruction c) {
        return this.insert(ih, c.getInstructionList());
    }
    
    public InstructionHandle insert(final InstructionHandle ih, final Instruction i) {
        return this.insert(ih, new InstructionList(i));
    }
    
    public InstructionHandle insert(final InstructionHandle ih, final InstructionList il) {
        if (il == null) {
            throw new ClassGenException("Inserting null InstructionList");
        }
        if (il.isEmpty()) {
            return ih;
        }
        final InstructionHandle prev = ih.getPrev();
        final InstructionHandle ret = il.start;
        ih.setPrev(il.end);
        il.end.setNext(ih);
        il.start.setPrev(prev);
        if (prev != null) {
            prev.setNext(il.start);
        }
        else {
            this.start = il.start;
        }
        this.length += il.length;
        il.clear();
        return ret;
    }
    
    public InstructionHandle insert(final InstructionList il) {
        if (this.isEmpty()) {
            this.append(il);
            return this.start;
        }
        return this.insert(this.start, il);
    }
    
    public boolean isEmpty() {
        return this.start == null;
    }
    
    @Override
    public Iterator<InstructionHandle> iterator() {
        return new Iterator<InstructionHandle>() {
            private InstructionHandle ih = InstructionList.this.start;
            
            @Override
            public boolean hasNext() {
                return this.ih != null;
            }
            
            @Override
            public InstructionHandle next() throws NoSuchElementException {
                if (this.ih == null) {
                    throw new NoSuchElementException();
                }
                final InstructionHandle i = this.ih;
                this.ih = this.ih.getNext();
                return i;
            }
            
            @Override
            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }
    
    public void move(final InstructionHandle ih, final InstructionHandle target) {
        this.move(ih, ih, target);
    }
    
    public void move(final InstructionHandle start, final InstructionHandle end, final InstructionHandle target) {
        if (start == null || end == null) {
            throw new ClassGenException("Invalid null handle: From " + start + " to " + end);
        }
        if (target == start || target == end) {
            throw new ClassGenException("Invalid range: From " + start + " to " + end + " contains target " + target);
        }
        for (InstructionHandle ih = start; ih != end.getNext(); ih = ih.getNext()) {
            if (ih == null) {
                throw new ClassGenException("Invalid range: From " + start + " to " + end);
            }
            if (ih == target) {
                throw new ClassGenException("Invalid range: From " + start + " to " + end + " contains target " + target);
            }
        }
        final InstructionHandle prev = start.getPrev();
        InstructionHandle next = end.getNext();
        if (prev != null) {
            prev.setNext(next);
        }
        else {
            this.start = next;
        }
        if (next != null) {
            next.setPrev(prev);
        }
        else {
            this.end = prev;
        }
        start.setPrev(end.setNext(null));
        if (target == null) {
            if (this.start != null) {
                this.start.setPrev(end);
            }
            end.setNext(this.start);
            this.start = start;
        }
        else {
            next = target.getNext();
            target.setNext(start);
            start.setPrev(target);
            end.setNext(next);
            if (next != null) {
                next.setPrev(end);
            }
            else {
                this.end = end;
            }
        }
    }
    
    public void redirectBranches(final InstructionHandle oldTarget, final InstructionHandle newTarget) {
        for (InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
            final Instruction i = ih.getInstruction();
            if (i instanceof BranchInstruction) {
                final BranchInstruction b = (BranchInstruction)i;
                final InstructionHandle target = b.getTarget();
                if (target == oldTarget) {
                    b.setTarget(newTarget);
                }
                if (b instanceof Select) {
                    final InstructionHandle[] targets = ((Select)b).getTargets();
                    for (int j = 0; j < targets.length; ++j) {
                        if (targets[j] == oldTarget) {
                            ((Select)b).setTarget(j, newTarget);
                        }
                    }
                }
            }
        }
    }
    
    public void redirectExceptionHandlers(final CodeExceptionGen[] exceptions, final InstructionHandle oldTarget, final InstructionHandle newTarget) {
        for (final CodeExceptionGen exception : exceptions) {
            if (exception.getStartPC() == oldTarget) {
                exception.setStartPC(newTarget);
            }
            if (exception.getEndPC() == oldTarget) {
                exception.setEndPC(newTarget);
            }
            if (exception.getHandlerPC() == oldTarget) {
                exception.setHandlerPC(newTarget);
            }
        }
    }
    
    public void redirectLocalVariables(final LocalVariableGen[] lg, final InstructionHandle oldTarget, final InstructionHandle newTarget) {
        for (final LocalVariableGen element : lg) {
            final InstructionHandle start = element.getStart();
            final InstructionHandle end = element.getEnd();
            if (start == oldTarget) {
                element.setStart(newTarget);
            }
            if (end == oldTarget) {
                element.setEnd(newTarget);
            }
        }
    }
    
    private void remove(final InstructionHandle prev, InstructionHandle next) throws TargetLostException {
        InstructionHandle first;
        InstructionHandle last;
        if (prev == null && next == null) {
            first = this.start;
            last = this.end;
            final InstructionHandle instructionHandle = null;
            this.end = instructionHandle;
            this.start = instructionHandle;
        }
        else {
            if (prev == null) {
                first = this.start;
                this.start = next;
            }
            else {
                first = prev.getNext();
                prev.setNext(next);
            }
            if (next == null) {
                last = this.end;
                this.end = prev;
            }
            else {
                last = next.getPrev();
                next.setPrev(prev);
            }
        }
        first.setPrev(null);
        last.setNext(null);
        final List<InstructionHandle> targetVec = new ArrayList<InstructionHandle>();
        for (InstructionHandle ih = first; ih != null; ih = ih.getNext()) {
            ih.getInstruction().dispose();
        }
        final StringBuilder buf = new StringBuilder("{ ");
        for (InstructionHandle ih2 = first; ih2 != null; ih2 = next) {
            next = ih2.getNext();
            --this.length;
            if (ih2.hasTargeters()) {
                targetVec.add(ih2);
                buf.append(ih2.toString(true)).append(" ");
                ih2.setNext(ih2.setPrev(null));
            }
            else {
                ih2.dispose();
            }
        }
        buf.append("}");
        if (!targetVec.isEmpty()) {
            throw new TargetLostException(targetVec.toArray(InstructionHandle.EMPTY_ARRAY), buf.toString());
        }
    }
    
    public void removeObserver(final InstructionListObserver o) {
        if (this.observers != null) {
            this.observers.remove(o);
        }
    }
    
    public void replaceConstantPool(final ConstantPoolGen oldCp, final ConstantPoolGen newCp) {
        for (InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
            final Instruction i = ih.getInstruction();
            if (i instanceof CPInstruction) {
                final CPInstruction ci = (CPInstruction)i;
                final Constant c = oldCp.getConstant(ci.getIndex());
                ci.setIndex(newCp.addConstant(c, oldCp));
            }
        }
    }
    
    public void setPositions() {
        this.setPositions(false);
    }
    
    public void setPositions(final boolean check) {
        int maxAdditionalBytes = 0;
        int additionalBytes = 0;
        int index = 0;
        int count = 0;
        final int[] pos = new int[this.length];
        if (check) {
            for (InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
                final Instruction i = ih.getInstruction();
                if (i instanceof BranchInstruction) {
                    Instruction inst = ((BranchInstruction)i).getTarget().getInstruction();
                    if (!this.contains(inst)) {
                        throw new ClassGenException("Branch target of " + Const.getOpcodeName(i.getOpcode()) + ":" + inst + " not in instruction list");
                    }
                    if (i instanceof Select) {
                        final InstructionHandle[] targets2;
                        final InstructionHandle[] targets = targets2 = ((Select)i).getTargets();
                        for (final InstructionHandle target : targets2) {
                            inst = target.getInstruction();
                            if (!this.contains(inst)) {
                                throw new ClassGenException("Branch target of " + Const.getOpcodeName(i.getOpcode()) + ":" + inst + " not in instruction list");
                            }
                        }
                    }
                    if (!(ih instanceof BranchHandle)) {
                        throw new ClassGenException("Branch instruction " + Const.getOpcodeName(i.getOpcode()) + ":" + inst + " not contained in BranchHandle.");
                    }
                }
            }
        }
        for (InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
            final Instruction i = ih.getInstruction();
            ih.setPosition(index);
            pos[count++] = index;
            switch (i.getOpcode()) {
                case 167:
                case 168: {
                    maxAdditionalBytes += 2;
                    break;
                }
                case 170:
                case 171: {
                    maxAdditionalBytes += 3;
                    break;
                }
            }
            index += i.getLength();
        }
        for (InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
            additionalBytes += ih.updatePosition(additionalBytes, maxAdditionalBytes);
        }
        count = (index = 0);
        for (InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
            final Instruction i = ih.getInstruction();
            ih.setPosition(index);
            pos[count++] = index;
            index += i.getLength();
        }
        System.arraycopy(pos, 0, this.bytePositions = new int[count], 0, count);
    }
    
    public int size() {
        return this.length;
    }
    
    @Override
    public String toString() {
        return this.toString(true);
    }
    
    public String toString(final boolean verbose) {
        final StringBuilder buf = new StringBuilder();
        for (InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
            buf.append(ih.toString(verbose)).append("\n");
        }
        return buf.toString();
    }
    
    public void update() {
        if (this.observers != null) {
            for (final InstructionListObserver observer : this.observers) {
                observer.notify(this);
            }
        }
    }
}
